@@ -1,13 +1,13 @@
-require File.expand_path('../../../spec_helper', __FILE__)
-require File.expand_path('../fixtures/classes', __FILE__)
-require File.expand_path('../shared/each_line', __FILE__)
-require File.expand_path('../shared/each_line_without_block', __FILE__)
+require_relative '../../spec_helper'
+require_relative 'fixtures/classes'
+require_relative 'shared/each_line'
+require_relative 'shared/each_line_without_block'
 
 describe "String#lines" do
-  it_behaves_like(:string_each_line, :lines)
+  it_behaves_like :string_each_line, :lines
 
   it "returns an array when no block given" do
-    ary = "hello world".send(@method, ' ')
+    ary = "hello world".lines(' ')
     ary.should == ["hello ", "world"]
   end
 
