@@ -1,10 +1,10 @@
-#define RUBY_VERSION "2.5.3"
-#define RUBY_RELEASE_DATE "2018-10-18"
-#define RUBY_PATCHLEVEL 105
+#define RUBY_VERSION "2.6.2"
+#define RUBY_RELEASE_DATE RUBY_RELEASE_YEAR_STR"-"RUBY_RELEASE_MONTH_STR"-"RUBY_RELEASE_DAY_STR
+#define RUBY_PATCHLEVEL 47
 
-#define RUBY_RELEASE_YEAR 2018
-#define RUBY_RELEASE_MONTH 10
-#define RUBY_RELEASE_DAY 18
+#define RUBY_RELEASE_YEAR 2019
+#define RUBY_RELEASE_MONTH 3
+#define RUBY_RELEASE_DAY 13
 
 #include "ruby/version.h"
 
@@ -60,11 +60,11 @@
 # define RUBY_REVISION_STR ""
 #endif
 
-# define RUBY_DESCRIPTION	    \
+# define RUBY_DESCRIPTION_WITH(opt) \
     "ruby "RUBY_VERSION		    \
     RUBY_PATCHLEVEL_STR		    \
     " ("RUBY_RELEASE_DATE	    \
-    RUBY_REVISION_STR") "	    \
+    RUBY_REVISION_STR")"opt" "	    \
     "["RUBY_PLATFORM"]"
 # define RUBY_COPYRIGHT		    \
     "ruby - Copyright (C) "	    \
